import numpy as np
from sympy import symbols, sin, sqrt, cos, exp, sqrt, pi, lambdify, Function
import pandas as pd
from typing import Callable
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

MAX_ITER = 1e5


def symb_grad_2d(f: Function) -> (Function, Function):
    return f.diff(x), f.diff(y)


def callable_grad_2d(f: Function) -> (Callable[[float, float], float],
                                      Callable[[float, float], float]):
    partialx, partialy = symb_grad_2d(f)
    return (lambdify([x, y], partialx), lambdify([x, y], partialy))


def callable_func(f: Function) -> Callable[[float, float], float]:
    return lambdify([x, y], f, modules='numpy')


def base_gd(f: Function, init_pt: list[float], lr: float) -> \
        pd.DataFrame:
    df = pd.DataFrame(columns=['x', 'y', 'Cost', 'NormGrad'])
    partialx, partialy = callable_grad_2d(f)

    x, y = init_pt

    iter = 0
    grad = np.array([partialx(x, y), partialy(x, y)])

    f_call = callable_func(f)

    while iter < MAX_ITER:
        if np.linalg.norm(grad) < 1e-6:
            break
        grad = np.array([partialx(x, y), partialy(x, y)])
        df.loc[iter] = [x, y, f_call(x, y), np.linalg.norm(grad)]

        step = lr * grad

        x -= step[0]
        y -= step[1]

        iter += 1

    return df


def momentum_gd(f: Function, init_pt: list[float], lr: float, momentum: float)\
        -> pd.DataFrame:
    df = pd.DataFrame(columns=['x', 'y', 'Cost', 'NormGrad'])
    partialx, partialy = callable_grad_2d(f)

    x, y = init_pt

    iter = 0
    grad = np.array([partialx(x, y), partialy(x, y)])

    f_call = callable_func(f)

    step = np.array([0, 0])

    while iter < MAX_ITER:
        if np.linalg.norm(grad) < 1e-6:
            break

        grad = np.array([partialx(x, y), partialy(x, y)])
        df.loc[iter] = [x, y, f_call(x, y), np.linalg.norm(grad)]

        step = momentum * step + lr * grad

        x -= step[0]
        y -= step[1]

        iter += 1

    return df


def nesterov_gd(f: Function, init_pt: list[float], lr: float, momentum: float)\
        -> pd.DataFrame:
    df = pd.DataFrame(columns=['x', 'y', 'Cost', 'NormGrad'])
    partialx, partialy = callable_grad_2d(f)

    x, y = init_pt

    iter = 0
    grad = np.array([partialx(x, y), partialy(x, y)])

    f_call = callable_func(f)

    step = np.array([0, 0])

    while iter < MAX_ITER:
        grad = np.array([partialx(x, y), partialy(x, y)])
        if np.linalg.norm(grad) < 1e-6:
            break

        grad_with_prev_step = np.array([partialx(x - step[0], y - step[1]),
                                        partialy(x - step[0], y - step[1])])
        df.loc[iter] = [x, y, f_call(x, y), np.linalg.norm(grad)]

        step = momentum * step + lr * grad_with_prev_step

        x -= step[0]
        y -= step[1]

        iter += 1

    return df


def adam_gd(f: Function, init_pt: list[float], lr: float)\
        -> pd.DataFrame:
    df = pd.DataFrame(columns=['x', 'y', 'Cost', 'NormGrad'])
    partialx, partialy = callable_grad_2d(f)

    f_call = callable_func(f)

    beta1 = 0.9
    beta2 = 0.999

    m = np.array([0, 0])
    v = np.array([0, 0])

    x, y = init_pt

    iter = 0
    grad = np.array([partialx(x, y), partialy(x, y)])

    while iter < MAX_ITER:
        grad = np.array([partialx(x, y), partialy(x, y)])
        if np.linalg.norm(grad) < 1e-6:
            break

        m = beta1 * m + (1 - beta1) * grad
        v = beta2 * v + (1 - beta2) * np.square(grad)

        if iter == 0:
            m_hat = np.array([0, 0])
            v_hat = np.array([0, 0])
        else:
            m_hat = m / (1 - beta1**iter)
            v_hat = v / (1 - beta2**iter)

        eta = lr / (np.sqrt(v_hat) + 1e-8)

        df.loc[iter] = [x, y, f_call(x, y), np.linalg.norm(grad)]

        step = eta * m_hat

        x -= step[0]
        y -= step[1]

        iter += 1

    return df


if __name__ == "__main__":
    x, y = symbols('x y')

    f: Function = x**2 + 7 * y**2
    # f: Function = 1 - exp(-10 * x**2 - y**2)
    # f: Function = x**2 * y - 2 * x * y**3 + 3 * x * y + 4

    # Rosenbrock(x, y)
    # f: Function = (1 - x)**2 + 1 * (y - x**2)**2

    # Beale(x, y)
    # f: Function = (1.5 - x + x * y)**2 + (2.25 - x + x *
    #                                       y**2)**2 + (2.625 - x + x * y**3)**2

    # Booth(x, y)
    # f: Function = (x + 2 * y - 7)**2 + (2 * x + y - 5)**2

    # Ackley(x, y)
    # f: Function = -20.0 * exp(-0.2 * sqrt(0.5 * (x**2 + y**2))) - \
    #     exp(0.5 * (cos(2 * pi * x) + cos(2 * pi * y))) + exp(1) + 20

    f_call = callable_func(f)

    LR = 1e-2
    MOMENTUM = 0.9

    plot_range = (7, 7)

    init_pt = np.array([np.random.randint(-plot_range[0], plot_range[0] + 1),
                        np.random.randint(-plot_range[1], plot_range[1] + 1)])

    base = base_gd(f, init_pt, LR)
    momentum = momentum_gd(f, init_pt, LR, MOMENTUM)
    nesterov = nesterov_gd(f, init_pt, LR, MOMENTUM)
    adam = adam_gd(f, init_pt, 0.7)

    print(f"Base = {base.tail(1)}")
    print(f"Momentum = {momentum.tail(1)}")
    print(f"Nesterov = {nesterov.tail(1)}")
    print(f"Adam = {adam.tail(1)}")

    x = np.linspace(-plot_range[0], plot_range[0], 100)
    y = np.linspace(-plot_range[1], plot_range[1], 100)

    X, Y = np.meshgrid(x, y)
    Z = f_call(X, Y)

    fig = plt.figure(1)
    ax = plt.axes(projection='3d')
    ax.plot_surface(X, Y, Z, cmap='jet', rstride=1,
                    cstride=1, norm=LogNorm(), alpha=0.4)
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')

    ax.contour(X, Y, Z, 20, cmap='RdGy', offset=0)

    ax.plot(base['x'], base['y'], '-o', color='red', label='Base')
    ax.plot(momentum['x'], momentum['y'], '-o',
            color='cyan', label='Momentum')
    ax.plot(nesterov['x'], nesterov['y'], '-o',
            color='orange', label='Nesterov')
    plt.plot(adam['x'], adam['y'], '-o',
             color='violet', label='Adam')
    ax.legend()

    plt.figure(2)
    plt.contour(X, Y, Z, 20, cmap='RdGy')
    plt.plot(base['x'], base['y'], '-o', color='red', label='Base')
    plt.plot(momentum['x'], momentum['y'], '-o',
             color='cyan', label='Momentum')
    plt.plot(nesterov['x'], nesterov['y'], '-o',
             color='orange', label='Nesterov')
    plt.plot(adam['x'], adam['y'], '-o',
             color='violet', label='Adam')
    plt.legend()
    plt.xlabel('x')
    plt.ylabel('y')

    plt.show()
